"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
const aws_sdk_1 = require("aws-sdk");
const fs = require("fs");
const Path = require("path");
const util_1 = require("util");
const AwsUtils_1 = require("./AwsUtils");
const Network_1 = require("./Network");
const ServerlessObjUtils_1 = require("./ServerlessObjUtils");
const SetupRepo_1 = require("./SetupRepo");
const Sleep_1 = require("./Sleep");
const readFile = util_1.promisify(fs.readFile);
const deepEqual = require("deep-equal");
class Plugin {
    constructor(serverless) {
        this.serverless = serverless;
        this.hooks = {
            "before:aws:deploy:deploy:updateStack": this.validate.bind(this),
            "after:aws:deploy:deploy:updateStack": this.setupElasticCache.bind(this)
        };
    }
    validate() {
        return __awaiter(this, void 0, void 0, function* () {
            const custom = this.serverless.service.custom || {};
            const configs = [].concat(custom.elasticsearch || {});
            for (const config of configs) {
                if (!config.endpoint && !config["cf-endpoint"]) {
                    throw new Error("Elasticsearch endpoint not specified.");
                }
            }
        });
    }
    setupElasticCache() {
        return __awaiter(this, void 0, void 0, function* () {
            const serviceName = yield ServerlessObjUtils_1.getProviderName(this.serverless);
            const profile = ServerlessObjUtils_1.getProfile(this.serverless);
            const region = ServerlessObjUtils_1.getRegion(this.serverless);
            const requestOptions = {};
            const credentials = {};
            const sts = new aws_sdk_1.STS({
                credentials: profile ? new aws_sdk_1.SharedIniFileCredentials({ profile }) : undefined,
                region,
                endpoint: AwsUtils_1.getServiceUrl("sts", region)
            });
            if (serviceName === "aws") {
                const creds = yield AwsUtils_1.assumeRole(sts, profile);
                aws_sdk_1.config.credentials = creds;
                aws_sdk_1.config.region = region;
                credentials.accessKeyId = creds.accessKeyId;
                credentials.secretAccessKey = creds.secretAccessKey;
                credentials.sessionToken = creds.sessionToken;
                credentials.region = region;
                credentials.service = "es";
            }
            const configs = yield parseConfig(this.serverless);
            const templateVariables = { indices: {} };
            for (const config of configs) {
                if (config.hasOwnProperty("onlyOnRegion") && config.onlyOnRegion !== region) {
                    this.serverless.cli.log(`Set to only run on region ${config.onlyOnRegion}, but deploying to region ${region}. Skipping...`);
                    continue;
                }
                const endpoint = config.endpoint.startsWith("http") ? config.endpoint : `https://${config.endpoint}`;
                this.serverless.cli.log(`Settings up endpoint ${endpoint}.`);
                this.serverless.cli.log("Setting up ingestion pipelines...");
                yield setupIngestionPipelines(endpoint, config.pipelines, requestOptions, credentials);
                this.serverless.cli.log("Setting up templates...");
                const setupTemplateResult = yield setupTemplates(endpoint, config.templates, { cli: this.serverless.cli }, requestOptions, credentials);
                for (const setupTemplate of setupTemplateResult.templates) {
                    const { swaps } = setupTemplate;
                    for (const swap of swaps) {
                        templateVariables.indices[stripVersion(swap.oldIndex)] = swap.newIndex;
                    }
                }
                this.serverless.cli.log("Setting up indices...");
                yield setupIndices(endpoint, config.indices, requestOptions, credentials);
                this.serverless.cli.log("Setting up repositories...");
                yield SetupRepo_1.setupRepo({
                    baseUrl: endpoint,
                    sts,
                    repos: config.repositories,
                    requestOptions,
                    credentials
                });
            }
            this.serverless.cli.log("Elasticsearch setup complete.");
        });
    }
}
function replaceServerlessTemplates(serverlessVariables, objToReplace) {
    const keyRegex = /{{esSetup\.([a-z._\-0-9]+):([ a-z0-9]+)}}/;
    if (typeof objToReplace === "string") {
        const match = objToReplace.match(keyRegex);
        if (!!match) {
            const variableName = match[1];
            const variableValue = match[2];
            const serverlessVariableKeys = `${variableName}.${variableValue}`.split(".").map(s => s.trim());
            let serverlessVariableValue = serverlessVariables;
            let wasSet = false;
            for (const key of serverlessVariableKeys) {
                wasSet = serverlessVariableValue.hasOwnProperty(key);
                if (wasSet) {
                    serverlessVariableValue = serverlessVariableValue[key];
                }
            }
            return wasSet ? objToReplace.replace(match[0], serverlessVariableValue) : objToReplace;
        }
    }
    else if (Array.isArray(objToReplace)) {
        for (let i = 0; i < objToReplace.length; ++i) {
            objToReplace[i] = replaceServerlessTemplates(serverlessVariables, objToReplace[i]);
        }
    }
    else if (typeof objToReplace === "object") {
        const serverlessKeys = Object.keys(objToReplace || {});
        for (const serverlessKey of serverlessKeys) {
            objToReplace[serverlessKey] = replaceServerlessTemplates(serverlessVariables, objToReplace[serverlessKey]);
        }
    }
    return objToReplace;
}
exports.replaceServerlessTemplates = replaceServerlessTemplates;
function parseConfig(serverless) {
    return __awaiter(this, void 0, void 0, function* () {
        const provider = serverless.service.provider;
        const custom = serverless.service.custom || {};
        let configs = [].concat(custom.elasticsearch || {});
        const returnConfigs = [];
        for (const config of configs) {
            if (config.hasOwnProperty("onlyOnRegion") && config.onlyOnRegion !== ServerlessObjUtils_1.getRegion(serverless)) {
                continue;
            }
            if (provider.name === "aws" || config["cf-endpoint"]) {
                const proifle = ServerlessObjUtils_1.getProfile(serverless);
                const cloudFormation = new aws_sdk_1.CloudFormation({
                    credentials: proifle ? new aws_sdk_1.SharedIniFileCredentials({ profile: ServerlessObjUtils_1.getProfile(serverless) }) : undefined,
                    endpoint: AwsUtils_1.getServiceUrl("cloudformation", ServerlessObjUtils_1.getRegion(serverless))
                });
                const newConfig = yield AwsUtils_1.parseConfigObject(cloudFormation, ServerlessObjUtils_1.getStackName(serverless), config);
                if (newConfig["cf-endpoint"]) {
                    newConfig.endpoint = yield AwsUtils_1.findCloudformationExport(cloudFormation, config["cf-endpoint"]);
                    if (!newConfig.endpoint) {
                        throw new Error("Endpoint not found at cloudformation export.");
                    }
                }
                returnConfigs.push(newConfig);
            }
        }
        return returnConfigs;
    });
}
function setupIndices(baseUrl, indices = [], requestOptions, credentials) {
    const setupPromises = indices.map((index) => {
        validateIndex(index);
        const url = `${baseUrl}/${index.name}`;
        const settings = require(Path.resolve(index.file));
        return Network_1.esPut(url, settings, requestOptions, credentials).catch((e) => {
            if (e.error.error.type !== "resource_already_exists_exception") {
                throw e;
            }
        });
    });
    return Promise.all(setupPromises);
}
function validateIndex(index) {
    if (!index.name) {
        throw new Error("Index does not have a name.");
    }
    if (!index.file) {
        throw new Error("Index does not have a file location.");
    }
}
function setupTemplates(baseUrl, templates = [], opts = {}, requestOptions, credentials) {
    return __awaiter(this, void 0, void 0, function* () {
        const { cli = { log: () => { } } } = opts;
        const setupPromises = templates.map((template) => __awaiter(this, void 0, void 0, function* () {
            validateTemplate(template);
            const url = `${baseUrl}/_template/${template.name}`;
            const file = yield readFile(Path.resolve(template.file)).then((file) => file.toString("utf-8"));
            const settings = JSON.parse(swapTemplateParameters(template, file));
            const returnValue = Object.assign(Object.assign({}, template), { swaps: [] });
            const { shouldSwapIndicesOfAliases } = template;
            if (!!template.shouldSwapIndicesOfAliases) {
                const reIndexPipeline = typeof shouldSwapIndicesOfAliases === "object" ? shouldSwapIndicesOfAliases.reIndexPipeline : undefined;
                const previous = yield returnPreviousTemplates(baseUrl, template.name, requestOptions, credentials);
                if (!!previous) {
                    const _a = previous[template.name], { order } = _a, previousSettings = __rest(_a, ["order"]);
                    if (!deepEqual(previousSettings.mappings, settings.mappings)) {
                        const { aliases } = previousSettings;
                        if (!!aliases) {
                            const aliasNames = Object.keys(aliases);
                            for (const aliasName of aliasNames) {
                                const swapResult = yield swapIndicesOfAliases({ cli, aliasName, baseUrl, reIndexPipeline }, requestOptions, credentials);
                                returnValue.swaps.push(...swapResult.swaps);
                            }
                        }
                    }
                }
            }
            return Network_1.esPut(url, settings, requestOptions, credentials)
                .then(() => returnValue);
        }));
        return Promise.all(setupPromises)
            .then((results) => results.reduce((returnValue, current) => {
            returnValue.templates.push(current);
            return returnValue;
        }, { templates: [] }));
    });
}
function returnPreviousTemplates(baseUrl, templateName, requestOptions, credentials) {
    const url = `${baseUrl}/_template/${templateName}`;
    return Network_1.esGet(url, undefined, requestOptions, credentials)
        .then(result => JSON.parse(result))
        .catch((error) => {
        if (error.statusCode === 404) {
            return undefined;
        }
        throw error;
    });
}
function returnCurrentAliases(baseUrl, aliasName, requestOptions, credentials) {
    const url = `${baseUrl}/_alias/${aliasName}`;
    return Network_1.esGet(url, undefined, requestOptions, credentials)
        .then(result => JSON.parse(result))
        .catch((error) => {
        if (error.statusCode === 404) {
            return undefined;
        }
        throw error;
    });
}
function validateTemplate(template) {
    if (!template.name) {
        throw new Error("Template does not have a name.");
    }
    if (!template.file) {
        throw new Error("Template does not have a file location.");
    }
}
function swapTemplateParameters(template, fileContent) {
    const { parameters = {} } = template;
    let replacedContent = fileContent;
    for (const paramKey of Object.keys(parameters)) {
        const regex = new RegExp(`\\\${${paramKey}}`, "g");
        replacedContent = replacedContent.replace(regex, parameters[paramKey]);
    }
    return replacedContent;
}
function swapIndicesOfAliases(props, requestOptions, credentials) {
    return __awaiter(this, void 0, void 0, function* () {
        const { cli = { log: () => { } }, baseUrl, aliasName } = props;
        cli.log(`Retrieving indices for ${aliasName}.`);
        const currentAliases = yield returnCurrentAliases(baseUrl, aliasName, requestOptions, credentials);
        const returnValue = {
            swaps: []
        };
        const currentIndices = !!currentAliases ? Object.keys(currentAliases) : [];
        for (const currentIndex of currentIndices) {
            const newIndex = incrementVersionValue(currentIndex);
            cli.log(`Creating index ${newIndex}.`);
            const createIndexUrl = `${baseUrl}/${newIndex}`;
            yield Network_1.esPut(createIndexUrl, {}, requestOptions, credentials)
                .catch((error) => {
                cli.log(`Failed to create index ${newIndex}: ${error.message}`);
                throw error;
            });
            cli.log(`Reindexing ${currentIndex} to ${newIndex}.`);
            const reindexUrl = `${baseUrl}/_reindex?wait_for_completion=false`;
            const reindexBody = {
                source: {
                    index: currentIndex
                },
                dest: {
                    index: newIndex,
                    pipeline: props.reIndexPipeline
                }
            };
            const response = yield Network_1.esPost(reindexUrl, reindexBody, requestOptions, credentials)
                .catch((error) => {
                cli.log(`Failed to reindex ${currentIndex} to ${newIndex}: ${error.message}`);
                throw error;
            });
            const reindexTaskToken = response.task;
            cli.log("Waiting for reindex task to complete.");
            yield waitForTaskCompletion({
                baseUrl,
                cli,
                taskId: reindexTaskToken,
            }, requestOptions, credentials);
            cli.log(`Swapping ${currentIndex} to ${newIndex} on alias ${aliasName}.`);
            const aliasSwapUrl = `${baseUrl}/_aliases`;
            const aliasSwapBody = {
                actions: [{
                        add: {
                            index: newIndex,
                            alias: stripVersion(currentIndex)
                        }
                    },
                    {
                        add: {
                            index: newIndex,
                            alias: aliasName
                        }
                    },
                    {
                        remove_index: {
                            index: currentIndex
                        }
                    }]
            };
            yield Network_1.esPost(aliasSwapUrl, aliasSwapBody, requestOptions, credentials)
                .catch((e) => {
                cli.log(`Failed to swap indices: ${e.message}`);
                throw e;
            });
            returnValue.swaps.push({
                alias: aliasName,
                oldIndex: currentIndex,
                newIndex
            });
        }
        return returnValue;
    });
}
function waitForTaskCompletion(props, requestOptions, credentials) {
    return __awaiter(this, void 0, void 0, function* () {
        const { baseUrl, taskId, cli } = props;
        let sleepTime = 15;
        const url = `${baseUrl}/_tasks`;
        const nodeAndTask = taskId.split(":");
        const task = yield Network_1.esGet(url, undefined, requestOptions, credentials)
            .then(result => JSON.parse(result))
            .then(tasks => tasks.nodes[nodeAndTask[0]].tasks[taskId])
            .catch((e) => {
            if (e.statusCode === 404) {
                return { completed: true };
            }
            if (e.statusCode === 429) {
                sleepTime = 60 * 5;
            }
            throw e;
        });
        if (!task || task.completed) {
            return;
        }
        cli.log(`Waiting for task ${taskId} to complete.`);
        yield Sleep_1.sleep(sleepTime);
        yield waitForTaskCompletion(props, requestOptions, credentials);
    });
}
function incrementVersionValue(value) {
    const versionRegex = /(.+)(_v(\d+)?)/;
    const matches = value.match(versionRegex);
    const name = !!matches ? matches[1] : value;
    const versionNumber = !!matches && !Number.isNaN(Number.parseInt(matches[3])) ? Number.parseInt(matches[3]) : 0;
    return `${name}_v${versionNumber + 1}`;
}
exports.incrementVersionValue = incrementVersionValue;
function stripVersion(value) {
    const versionRegex = /(.+)(_v(\d+)?)/;
    const matches = value.match(versionRegex);
    return !!matches ? matches[1] : value;
}
exports.stripVersion = stripVersion;
function setupIngestionPipelines(baseUrl, templates = [], requestOptions, credentials) {
    return __awaiter(this, void 0, void 0, function* () {
        const setupPromises = templates.map((template) => __awaiter(this, void 0, void 0, function* () {
            validateIngestionPipelineTemplate(template);
            const url = `${baseUrl}/_ingest/pipeline/${template.name}`;
            const file = yield readFile(Path.resolve(template.file)).then((file) => file.toString("utf-8"));
            const settings = JSON.parse(swapTemplateParameters(template, file));
            return Network_1.esPut(url, settings, requestOptions, credentials);
        }));
        return Promise.all(setupPromises);
    });
}
function validateIngestionPipelineTemplate(template) {
    if (!template.name) {
        throw new Error("Ingestion pipeline template does not have a name.");
    }
    if (!template.file) {
        throw new Error("Ingestion pipeline template does not have a file location.");
    }
}
exports.default = Plugin;
